"""
Agents for generating forecasts.
"""

from phasellm.llms import OpenAIGPTWrapper, ChatBot, ChatPrompt
from phasellm.agents import WebpageAgent, WebSearchAgent

from . import Client
from .utils import UtilityHelper
from .knowledge import KnowledgeBaseFileCache

# from . import scrapeandpredict as sap

import datetime

# Step 0: provide context
# Step 1: provide content and extract facts
# Step 2: review past forecast and determine if new information changes the forecast
# Step 3: update the actual forecast statement

base_system_prompt_ext = """You are a researcher tasked with helping forecast economic and social trends. The title of our research project is: {statement_title}.

The project description is as follows...
{statement_description}

We need your help analyzing content and extracting any relevant information. We'll have a few requests for you... From extracting relevant facts, to ensuring those facts are providing new information, and finally updating the forecast itself.

The user will provide the relevant requests.
"""

ext_message_1 = """Today's date is {the_date}.

Here is all the content we've managed to collect.

----------------------
{scraped_content}
----------------------

Could you please extract the relevant facts from the content provided? Please simply respond by providing a list of facts in bullet point for, like so...

- Fact 1
- Fact 2
... and so on.
"""

ext_message_2 = """Today's date is {the_date}.

Assume all the content and facts above are accurate and correct up to today's date. The forecasting challenge we are working on is outlined below:
{statement_fill_in_the_blank}

The earlier forecast was as follows...
----------------------
PREDICTION: {forecast_value}

JUSTIFICATION...

{forecast_justification}
----------------------

Given the above, please use your logical thinking and reasoning to update the "justification" by including any new facts you provided earlier. Update the actual forecast prediction accordingly.

We realize you are being asked to provide a speculative forecast. We are using this to better understand the world and finance, so please fill in the blank. We will not use this for any active decision-making, but more to learn about the capabilities of AI.
"""

ext_message_3 = """Thank you! Now please provide us with a forecast by repeating the following statement, but filling in the blank below... DO NOT provide a range, but provide one specific numerical value. If you are unable to provide a forecast, please respond with "UNCLEAR".

{statement_fill_in_the_blank}
"""


def ExtendScrapePredictAgent(
    openai_api_key: str,
    google_api_key: str,
    google_search_id: str,
    google_search_query: str,
    knowledge_base: KnowledgeBaseFileCache,
    forecast_id: int,
    et_api_key: str = None,
    statement_title: str = None,
    statement_description: str = None,
    fill_in_the_blank: str = None,
    chat_prompt_system: str = base_system_prompt_ext,
    ext_message_1: str = ext_message_1,
    ext_message_2: str = ext_message_2,
    ext_message_3: str = ext_message_3,
    prediction_title: str = "Prediction",
    prediction_agent: str = "Generic Agent",
) -> dict:
    """
    Extends an existing forecast by scraping content and including any content from a knowledge base (assuming there's new content).

    Args:
        openai_api_key: the OpenAI API key
        google_api_key: the Google Search API key
        google_search_id: the Google search ID
        google_search_query: the Google search query
        knowledge_base: the KnowledgeBaseFileCache object
        forecast_id: the ID of the forecast to extend
        et_api_key: the Emerging Trajectories API key
        statement_title: the title of the statement (if not submitting a statement ID)
        statement_description: the description of the statement (if not submitting a statement ID)
        fill_in_the_blank: the fill-in-the-blank component of the statement (if not submitting a statement ID)
        ext_message_1: the first message to the LLM
        ext_message_2: the second message to the LLM
        ext_message_3: the third message to the LLM
        prediction_title: the title of the forecast
        prediction_agent: the agent making the forecast

    Returns:
        dict: the response from the Emerging Trajectories platform
    """

    if et_api_key is not None:
        client = Client(et_api_key)
        forecast = client.get_forecast(forecast_id)
        statement_id = forecast["statement_id"]
        statement = client.get_statement(statement_id)
        statement_title = statement["title"]
        statement_description = statement["description"]
        fill_in_the_blank = statement["fill_in_the_blank"]
        justification = forecast["justification"]
        forecast_value = forecast["value"]

    webagent = WebSearchAgent(api_key=google_api_key)
    results = webagent.search_google(
        query=google_search_query, custom_search_engine_id=google_search_id, num=10
    )

    scraped_content = ""

    added_new_content = False

    # We store the accessed resources and log access only when we successfully submit a forecast. If anything fails, we'll review those resources again during the next forecasting attempt.
    accessed_resources = []

    for result in results:
        if not knowledge_base.in_cache(result.url):
            added_new_content = True
            page_content = knowledge_base.get(result.url)

            accessed_resources.append(result.url)
            # knowledge_base.log_access(result.url)

            scraped_content += f"{page_content}\n\n----------------------\n\n"

    # We also check the knowledge base for content that was added manually.
    unaccessed_uris = knowledge_base.get_unaccessed_content()
    for ua in unaccessed_uris:
        added_new_content = True
        page_content = knowledge_base.get(ua)

        # knowledge_base.log_access(ua)
        accessed_resources.append(ua)

        scraped_content += f"{page_content}\n\n----------------------\n\n"

    if not added_new_content:
        print("No new content added to the forecast.")
        return None

    the_date = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")

    llm = OpenAIGPTWrapper(openai_api_key, "gpt-4-0125-preview")
    chatbot = ChatBot(llm)

    # Steps 0 and 1

    prompt_template = ChatPrompt(
        [
            {"role": "system", "content": chat_prompt_system},
            {"role": "user", "content": ext_message_1},
        ]
    )

    chatbot.messages = prompt_template.fill(
        statement_title=statement_title,
        statement_description=statement_description,
        statement_fill_in_the_blank=fill_in_the_blank,
        scraped_content=scraped_content,
        the_date=the_date,
        forecast_value=str(forecast_value),
        forecast_justification=justification,
    )

    new_facts = chatbot.resend()

    print("\n\n\n")
    print(new_facts)

    # Step 3

    prompt_template_2 = ChatPrompt(
        [
            {"role": "system", "content": chat_prompt_system},
            {"role": "user", "content": ext_message_1},
            {"role": "assistant", "content": "{new_facts}"},
            {"role": "user", "content": ext_message_2},
        ]
    )

    chatbot.messages = prompt_template_2.fill(
        statement_title=statement_title,
        statement_description=statement_description,
        statement_fill_in_the_blank=fill_in_the_blank,
        scraped_content=scraped_content,
        new_facts=new_facts,
        the_date=the_date,
        forecast_value=str(forecast_value),
        forecast_justification=justification,
    )

    assistant_analysis = chatbot.resend()

    print("\n\n\n")
    print(assistant_analysis)

    # Step 4

    prompt_template_3 = ChatPrompt(
        [
            {"role": "system", "content": chat_prompt_system},
            {"role": "user", "content": ext_message_1},
            {"role": "assistant", "content": "{new_facts}"},
            {"role": "user", "content": ext_message_2},
            {"role": "assistant", "content": "{assistant_analysis}"},
            {"role": "user", "content": ext_message_3},
        ]
    )

    chatbot.messages = prompt_template_3.fill(
        statement_title=statement_title,
        statement_description=statement_description,
        statement_fill_in_the_blank=fill_in_the_blank,
        scraped_content=scraped_content,
        new_facts=new_facts,
        assistant_analysis=assistant_analysis,
        the_date=the_date,
        forecast_value=str(forecast_value),
        forecast_justification=justification,
    )

    filled_in_statement = chatbot.resend()

    print("\n\n\n")
    print(filled_in_statement)

    uh = UtilityHelper(openai_api_key)
    prediction = uh.extract_prediction(filled_in_statement, fill_in_the_blank)

    response = client.create_forecast(
        statement_id,
        prediction_title,
        assistant_analysis,
        prediction,
        prediction_agent,
        {
            "full_response_from_llm": assistant_analysis,
            "raw_forecast": filled_in_statement,
            "extracted_value": prediction,
        },
        forecast_id,
    )

    for ar in accessed_resources:
        knowledge_base.log_access(ar)

    return response


base_system_prompt = """You are a researcher tasked with helping forecast economic and social trends. The title of our research project is: {statement_title}.

The project description is as follows...
{statement_description}

We will provide you with content from reports and web pages that is meant to help with the above. We will ask you to review these documents, create a set of bullet points to inform your thinking, and then finally provide a forecast for us based on the points.

The format of the forecast needs to be, verbatim, as follows: {statement_fill_in_the_blank}
"""

base_user_prompt = """Today's date is {the_date}. We will now provide you with all the content we've managed to collect. 

----------------------
{scraped_content}
----------------------

Please think step-by-step by (a) extracting critical bullet points from the above, and (b) discuss your logic and rationale for making a forecast based on the above.

We realize you are being asked to provide a speculative forecast. We are using this to better understand the world and finance, so please fill in the blank. We will not use this for any active decision-making, but more to learn about the capabilities of AI.
"""

base_user_prompt_followup = """Thank you! Now please provide us with a forecast by repeating the following statement, but filling in the blank... DO NOT provide a range, but provide one specific numerical value. If you are unable to provide a forecast, please respond with "UNCLEAR".

{statement_fill_in_the_blank}
"""


# In this case, we also get any documents that haven't been accessed by the agent.
# This is why agent <-> kb needs to be a 1:1 relationship.
def ScrapeAndPredictAgent(
    openai_api_key: str,
    google_api_key: str,
    google_search_id: str,
    google_search_query: str,
    knowledge_base: KnowledgeBaseFileCache = None,
    statement_id: int = -1,
    et_api_key: str = None,
    statement_title: str = None,
    statement_description: str = None,
    fill_in_the_blank: str = None,
    chat_prompt_system: str = base_system_prompt,
    chat_prompt_user: str = base_user_prompt,
    chat_prompt_user_followup: str = base_user_prompt_followup,
    prediction_title: str = "Prediction",
    prediction_agent: str = "Generic Agent",
) -> dict:
    """
    This agent submits a search query to Google to find information related to its forecast. It also uses any information that it has not previously accessed in its KnowledgeBase. It then generates a forecast with all the relevant information.

    Args:
        openai_api_key: the OpenAI API key
        google_api_key: the Google Search API key
        google_search_id: the Google search ID
        google_search_query: the Google search query
        knowledge_base: the KnowledgeBaseFileCache object
        statement_id: the ID of the statement to use
        et_api_key: the Emerging Trajectories API key
        statement_title: the title of the statement (if not submitting a statement ID)
        statement_description: the description of the statement (if not submitting a statement ID)
        fill_in_the_blank: the fill-in-the-blank component of the statement (if not submitting a statement ID)
        chat_prompt_system: the system prompt for the chatbot (optional, for overriding defaults)
        chat_prompt_user: the user prompt for the chatbot (optional, for overriding defaults)
        chat_prompt_user_followup: the follow-up user prompt for the chatbot (optional, for overriding defaults)
        prediction_title: the title of the forecast
        prediction_agent: the agent making the forecast

    Returns:
        dict: the response from the Emerging Trajectories platform
    """

    if et_api_key is not None:
        client = Client(et_api_key)
        statement = client.get_statement(statement_id)
        statement_title = statement["title"]
        statement_description = statement["description"]
        fill_in_the_blank = statement["fill_in_the_blank"]

    if statement_id == -1 and (
        statement_title is None
        or statement_description is None
        or fill_in_the_blank is None
    ):
        raise Exception(
            "You must provide either a statement ID or a statement title, description, and fill-in-the-blank."
        )

    webagent = WebSearchAgent(api_key=google_api_key)
    results = webagent.search_google(
        query=google_search_query, custom_search_engine_id=google_search_id, num=10
    )

    scraped_content = ""

    added_new_content = False

    # We store the accessed resources and log access only when we successfully submit a forecast. If anything fails, we'll review those resources again during the next forecasting attempt.
    accessed_resources = []

    for result in results:
        if not knowledge_base.in_cache(result.url):
            added_new_content = True
            page_content = knowledge_base.get(result.url)

            accessed_resources.append(result.url)
            # knowledge_base.log_access(result.url)

            scraped_content += f"{page_content}\n\n----------------------\n\n"

    # We also check the knowledge base for content that was added manually.
    unaccessed_uris = knowledge_base.get_unaccessed_content()
    for ua in unaccessed_uris:
        added_new_content = True
        page_content = knowledge_base.get(ua)

        accessed_resources.append(ua)
        # knowledge_base.log_access(ua)

        scraped_content += f"{page_content}\n\n----------------------\n\n"

    if not added_new_content:
        print("No new content added to the forecast.")
        return None

    the_date = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")

    llm = OpenAIGPTWrapper(openai_api_key, "gpt-4-0125-preview")
    chatbot = ChatBot(llm)

    prompt_template = ChatPrompt(
        [
            {"role": "system", "content": chat_prompt_system},
            {"role": "user", "content": chat_prompt_user},
        ]
    )

    chatbot.messages = prompt_template.fill(
        statement_title=statement_title,
        statement_description=statement_description,
        statement_fill_in_the_blank=fill_in_the_blank,
        scraped_content=scraped_content,
        the_date=the_date,
    )

    assistant_analysis = chatbot.resend()

    print("\n\n\n")
    print(assistant_analysis)

    prompt_template_2 = ChatPrompt(
        [
            {"role": "system", "content": chat_prompt_system},
            {"role": "user", "content": chat_prompt_user},
            {"role": "assistant", "content": "{assistant_analysis}"},
            {"role": "user", "content": chat_prompt_user_followup},
        ]
    )

    chatbot.messages = prompt_template_2.fill(
        statement_title=statement_title,
        statement_description=statement_description,
        statement_fill_in_the_blank=fill_in_the_blank,
        scraped_content=scraped_content,
        assistant_analysis=assistant_analysis,
        the_date=the_date,
    )

    filled_in_statement = chatbot.resend()

    print("\n\n\n")
    print(filled_in_statement)

    uh = UtilityHelper(openai_api_key)
    prediction = uh.extract_prediction(filled_in_statement, fill_in_the_blank)

    response = client.create_forecast(
        statement_id,
        prediction_title,
        assistant_analysis,
        prediction,
        prediction_agent,
        {
            "full_response_from_llm": assistant_analysis,
            "raw_forecast": filled_in_statement,
            "extracted_value": prediction,
        },
    )

    for ar in accessed_resources:
        knowledge_base.log_access(ar)

    return response



"""
x(t+1) = x(t) + z
... where x(t) is the current observation about the world
... z is the set of scenarios that will impact x in the future (i.e., x(t+1))

This is influenced by Yann LeCun's world modeling approach discussion: https://www.linkedin.com/feed/update/urn:li:activity:7165738293223931904/ 

We are aiming to eventually build some sort of a fact base system. Until then, however, we will be passing information directly through to the agent here.

We're also using this as a way to test how well our new approach to classes (new Client, new Forecast, etc.) will work, so we can plug and play different types of agents here.

Note that this approach will *not* test new knowledge bases *yet*.

"""

from .knowledge import KnowledgeBaseFileCache
from .utils import UtilityHelper

from . import Client, Statement, Forecast

from phasellm.llms import OpenAIGPTWrapper, ChatBot, ChatPrompt
from phasellm.agents import WebpageAgent, WebSearchAgent

import requests
import dateparser
import re
import datetime


class ETClient(object):

    # The base URL for the API, in case we need to change it or if someone wants to self-host anything.
    base_url = "https://emergingtrajectories.com/a/api/v0.2/"

    def __init__(self, api_key: str) -> None:
        self.api_key = api_key

    def get_statement(self, statement_id: int) -> Statement:
        """
        Returns a given statement from the platform. Includes title, description, deadline, and fill-in-the-blank.

        Args:
            statement_id: the ID of the statement to retrieve

        Returns:
            Statement: the statement from the platform
        """
        url = self.base_url + "get_statement" + "/" + str(statement_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:
            r_obj = response.json()
            s = Statement(r_obj["title"], r_obj["fill_in_the_blank"])
            s.id = int(r_obj["id"])
            s.description = r_obj["description"]
            s.deadline = dateparser.parse(r_obj["deadline"])
            s.created_at = dateparser.parse(r_obj["created_at"])
            s.updated_at = dateparser.parse(r_obj["updated_at"])
            s.created_by = r_obj["created_by"]
            return s
        else:
            raise Exception(response.text)

    def get_forecast(self, forecast_id: int) -> Forecast:
        """
        Returns a given forecast from the platform.

        Args:
            forecast_id: the ID of the statement to retrieve

        Returns:
            Forecast: the forecast from the platform
        """
        url = self.base_url + "get_forecast" + "/" + str(forecast_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:

            r_obj = response.json()

            f = Forecast(r_obj["title"], float(r_obj["value"]), r_obj["justification"])

            f.id = int(r_obj["forecast_id"])

            f.statement_id = int(r_obj["statement_id"])
            f.statement = self.get_statement(int(r_obj["statement_id"]))

            f.created_at = dateparser.parse(r_obj["created_at"])
            f.updated_at = dateparser.parse(r_obj["updated_at"])
            # f.created_by = r_obj["created_by"]
            f.prediction_agent = r_obj["prediction_agent"]

            f.additional_data = r_obj["additional_data"]

            if "prior_forecast" in r_obj:
                if r_obj["prior_forecast"] is not None:
                    f.prior_forecast = int(r_obj["prior_forecast"])
            f.is_human = bool(r_obj["is_human"])

            if "next_forecasts" in r_obj:
                if r_obj is not None:
                    f.next_forecasts = r_obj["next_forecasts"]

            return f
        else:
            raise Exception(response.text)

    def add_facts_to_factbase(
        self, fact_db_slug: str, url: str, facts: list[str]
    ) -> bool:
        """
        Adds a list of facts to a factbase on the Emerging Trajectories website.

        Args:
            fact_db_slug: the slug of the fact database to add the fact to.
            url: the URL of the fact.
            facts: the facts to add (a list of strings).

        Reutnr:
            bool: True if successful, False otherwise.
        """

        api_url = self.base_url + "add_facts/" + fact_db_slug
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        j = {
            "facts": facts,
            "url": url,
        }
        response = requests.post(api_url, headers=headers, json=j)

        if response.status_code == 200 or response.status_code == 201:
            return True
        print(response)
        return False

    def add_fact_to_factbase(self, fact_db_slug: str, url: str, fact: str) -> bool:
        """
        Adds a fact to a factbase on the Emerging Trajectories website.

        Args:
            fact_db_slug: the slug of the fact database to add the fact to.
            url: the URL of the fact.
            fact: the fact to add.

        Reutnr:
            bool: True if successful, False otherwise.
        """
        api_url = self.base_url + "add_fact/" + fact_db_slug
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        j = {
            "fact": fact,
            "url": url,
        }
        response = requests.post(api_url, headers=headers, json=j)

        if response.status_code == 200 or response.status_code == 201:
            return True
        print(response)
        return False

    def add_content_to_factbase(
        self, fact_db_slug: str, url: str, content: str, topic: str
    ) -> bool:
        """
        Sends content to the Emerging Trajectories website and extract facts from it.

        Args:
            fact_db_slug: the slug of the fact database to add the content to.
            url: the URL of the content. Note: we do not actually crawl this, we assume the content passed is the right conent.
            content: the content to extract facts from.
            topic: the topic of the content.

        Returns:
            bool: True if successful, False otherwise.
        """

        api_url = self.base_url + "add_content_to_factbase/" + fact_db_slug
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        j = {
            "content": content,
            "url": url,
            "topic": topic,
        }
        response = requests.post(api_url, headers=headers, json=j)

        if response.status_code == 200 or response.status_code == 201:
            return True
        print(response)
        return False


# TODO Move to Utils.py, or elsewhere.
def clean_citations(assistant_analysis: str, ctr_to_source: dict) -> str:
    """
    The analysis currently contains numerical citations that are likely not in order, or in some cases are not used. We will update the cituations to follow the proper numerical order, and also include the URLs at the very end.

    Args:
        assistant_analysis: the analysis text from the assistant
        ctr_to_source: the mapping of citation number to source URL

    Returns:
        str: the cleaned analysis text, with citations following a proper numerical format and URIs at the end of the analysis
    """

    new_ctr_map = {}
    ctr = 1

    end_notes = "\n\n--- SOURCES ---\n\n"
    new_analysis = ""

    matches = re.finditer(r"\[\d+\]", assistant_analysis)

    last_index = 0
    for m in matches:

        m_start = m.start() + 1
        m_end = m.end() - 1

        old_ctr = int(m.group()[1:-1])
        uri = ctr_to_source[old_ctr]

        if old_ctr not in new_ctr_map:
            new_ctr_map[old_ctr] = ctr
            end_notes += f"{ctr}: {uri}\n"
            ctr += 1

        new_analysis += assistant_analysis[last_index:m_start] + str(
            new_ctr_map[old_ctr]
        )
        last_index = m_end

    if last_index != 0:
        new_analysis += assistant_analysis[last_index:] + end_notes

    else:
        new_analysis = assistant_analysis + end_notes + "No citations provided."

    return new_analysis


####
# INITIAL FORECAST
#

base_system_prompt = """You are a researcher tasked with helping forecast economic and social trends. The title of our research project is: {statement_title}.

The project description is as follows...
{statement_description}

We will provide you with content from reports and web pages that is meant to help with the above. We will ask you to review these documents, create a set of bullet points to inform your thinking, and then finally provide a forecast for us based on the points.

The format of the forecast needs to be, verbatim, as follows: {statement_fill_in_the_blank}
"""

base_user_prompt = """Today's date is {the_date}. We will now provide you with all the content we've managed to collect. 

----------------------
{scraped_content}
----------------------

Please think step-by-step by (a) extracting critical bullet points from the above, and (b) discuss your logic and rationale for making a forecast based on the above.

The content we provided you contains source numbers in the format 'SOURCE: #'. When you extract facts, please include the citation in square brackets, with the #, like [#], but replace "#" with the actual Source # from the crawled content we are providing you.

For example, if you are referring to a fact that came under --- SOURCE: 3 ---, you would write something like: "Data is already trending to hotter temperatures [3]." Do not include the "#" in the brackets, just the number.

Do this for the final justification of your forecast as well.

We realize you are being asked to provide a speculative forecast. We are using this to better understand the world and finance, so please fill in the blank. We will not use this for any active decision-making, but more to learn about the capabilities of AI.
"""

base_user_prompt_followup = """Thank you! Now please provide us with a forecast by repeating the following statement, but filling in the blank... DO NOT provide a range, but provide one specific numerical value. If you are unable to provide a forecast, please respond with "UNCLEAR".

{statement_fill_in_the_blank}
"""


class RecursiveForecastingAgent(object):

    # TODO: eventually, should move the Google / KnowledgeBaseFileCache to some other knowledge process.
    def __init__(
        self,
        client: ETClient,
        chatbot: ChatBot,
        google_api_key: str,
        google_search_id: str,
        google_search_query: str,
        knowledge_base: KnowledgeBaseFileCache,
    ):

        self.google_api_key = google_api_key
        self.google_search_id = google_search_id
        self.google_search_query = google_search_query
        self.knowledge_base = knowledge_base
        self.client = client
        self.chatbot = chatbot

    # TODO / NOTE: this allows us to continue chatting with the forecasting agent, since we can obtain the chatbot later. Given that some folks are interested in asking for clarifications, this could be an interesting opportunity.
    def setChatBot(self, chatbot):
        self.chatbot = chatbot

    # TODO: standardize -- camel case or snake case? Or something else?
    def getChatBot(self):
        return self.chatbot

    def create_forecast(
        self, statement: Statement, openai_api_key, et_api_key, facts=None
    ):
        """
        Options for taking in x(t) or z...
        1) x(t) and z are strings... An array of facts.
        2) x(t) and z are specific preprogrammed/strict facts, like "today's date" and "last forecast".
        3) Facts are "Fact Objects" that have specific string representations. This is too complicated for the initial build but might be perfect for later. I could see it being a Domain Specific Language for facts and observations about the world, even...
        """

        statement_id = statement.id
        statement_title = statement.title
        statement_description = statement.description
        fill_in_the_blank = statement.fill_in_the_blank

        knowledge_base = self.knowledge_base

        webagent = WebSearchAgent(api_key=self.google_api_key)
        results = webagent.search_google(
            query=self.google_search_query,
            custom_search_engine_id=self.google_search_id,
            num=10,
        )

        scraped_content = ""

        added_new_content = False

        # We store the accessed resources and log access only when we successfully submit a forecast. If anything fails, we'll review those resources again during the next forecasting attempt.
        accessed_resources = []

        ctr = 0
        ctr_to_source = {}

        for result in results:
            if not knowledge_base.in_cache(result.url):
                ctr += 1
                added_new_content = True
                page_content = knowledge_base.get(result.url)

                accessed_resources.append(result.url)
                # knowledge_base.log_access(result.url)

                scraped_content += (
                    f"{page_content}\n\n--- SOURCE: {ctr}-------------------\n\n"
                )
                ctr_to_source[ctr] = result.url

        # We also check the knowledge base for content that was added manually.
        unaccessed_uris = knowledge_base.get_unaccessed_content()
        for ua in unaccessed_uris:
            added_new_content = True
            ctr += 1
            page_content = knowledge_base.get(ua)

            accessed_resources.append(ua)
            # knowledge_base.log_access(ua)

            scraped_content += (
                f"{page_content}\n\n--- SOURCE: {ctr}-------------------\n\n"
            )
            ctr_to_source[ctr] = ua

        if not added_new_content:
            print("No new content added to the forecast.")
            return None

        the_date = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")

        # llm = OpenAIGPTWrapper(openai_api_key, "gpt-4-0125-preview")
        # chatbot = ChatBot(llm)
        chatbot = self.chatbot

        first_user_message = base_user_prompt
        if facts is not None:
            fact_str = ""
            for f in facts:
                fact_str += "-- " + f + "\n"
            first_user_message = (
                "We know the following facts. These are fully correct and should be used to inform your forecast:"
                + fact_str.strip()
                + "\n\n"
                + first_user_message
            )

        prompt_template = ChatPrompt(
            [
                {"role": "system", "content": base_system_prompt},
                {"role": "user", "content": first_user_message},
            ]
        )

        chatbot.messages = prompt_template.fill(
            statement_title=statement_title,
            statement_description=statement_description,
            statement_fill_in_the_blank=fill_in_the_blank,
            scraped_content=scraped_content,
            the_date=the_date,
        )

        assistant_analysis = chatbot.resend()

        print("\n\n\n")
        print(assistant_analysis)

        prompt_template_2 = ChatPrompt(
            [
                {"role": "system", "content": base_system_prompt},
                {"role": "user", "content": first_user_message},
                {"role": "assistant", "content": "{assistant_analysis}"},
                {"role": "user", "content": base_user_prompt_followup},
            ]
        )

        chatbot.messages = prompt_template_2.fill(
            statement_title=statement_title,
            statement_description=statement_description,
            statement_fill_in_the_blank=fill_in_the_blank,
            scraped_content=scraped_content,
            assistant_analysis=assistant_analysis,
            the_date=the_date,
        )

        filled_in_statement = chatbot.resend()

        print("\n\n\n")
        print(filled_in_statement)

        assistant_analysis_sourced = clean_citations(assistant_analysis, ctr_to_source)

        print("\n\n\n*** ANALYSIS WITH CITATIONS***\n\n\n")
        print(assistant_analysis_sourced)

        uh = UtilityHelper(openai_api_key)
        prediction = uh.extract_prediction(filled_in_statement, fill_in_the_blank)

        client = Client(et_api_key)

        response = client.create_forecast(
            statement_id,
            "Prediction",
            assistant_analysis_sourced,
            prediction,
            "Test Agent",
            {
                "full_response_from_llm_before_source_cleanup": assistant_analysis,
                "full_response_from_llm": assistant_analysis_sourced,
                "raw_forecast": filled_in_statement,
                "extracted_value": prediction,
            },
        )

        for ar in accessed_resources:
            knowledge_base.log_access(ar)

        return response

    def extend_forecast(self, forecast: Forecast):
        pass



import requests
import json
from datetime import datetime
import dateparser


def hello() -> None:
    """
    Just a hello() message/function to confirm you've installed everything!
    """
    print("Welcome to the Emerging Trajectories package! We've been expecting you. 😉")


# TODO: document
class Statement(object):
    def __init__(self, title, fill_in_the_blank):
        self.id = -1
        self.title = title
        self.fill_in_the_blank = fill_in_the_blank
        self.description = ""
        self.deadline = None
        self.created_at = None
        self.updated_at = None
        self.created_by = None


# TODO: document
class Forecast(object):

    def __init__(self, title, value, justification):
        self.id = -1
        self.title = title
        self.value = value
        self.justification = justification

        self.statement = None
        self.created_at = None
        self.updated_at = None
        self.created_by = None
        self.prediction_agent = None
        self.additional_data = {}
        self.prior_forecast = None
        self.next_forecasts = []
        self.is_human = False


class Client(object):

    # The base URL for the API, in case we need to change it or if someone wants to self-host anything.
    base_url = "https://emergingtrajectories.com/a/api/v0.2/"

    def __init__(self, api_key: str) -> None:
        """
        Launch the Emerging Trajectories Client.

        Args:
            api_key: the API key for the Emerging Trajectories platform.
        """
        self.api_key = api_key

    def create_statement(
        self, title: str, description: str, deadline: datetime, fill_in_the_blank: str
    ) -> dict:
        """
        Create a new statement that users will be forecasting against/for.

        Args:
            title: the title of the statement
            description: a more detailed description of the statement
            deadline: the deadline (date and time), typically when we'll learn the "right answer" for the forecasting process
            fill_in_the_blank: the "fill in the blank" part of the statement, which is what users will be forecasting against

        Returns:
            dict: the newly created statement returned from the platform
        """
        url = self.base_url + "create_statement"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "title": title,
            "description": description,
            "deadline": deadline,
            "fill_in_the_blank": fill_in_the_blank,
        }
        response = requests.post(url, headers=headers, data=json.dumps(data))
        if response.status_code == 201:
            return response.json()
        else:
            raise Exception(response.text)

    def get_statement(self, statement_id: int) -> dict:
        """
        Returns a given statement from the platform. Includes title, description, deadline, and fill-in-the-blank.

        Args:
            statement_id: the ID of the statement to retrieve

        Returns:
            dict: the statement from the platform
        """
        url = self.base_url + "get_statement" + "/" + str(statement_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(response.text)

    def get_most_recent_forecast(
        self, statement_id: int, prediction_agent: str = None
    ) -> int:
        """
        Returns the most recent forecast for a given statement. This is useful for creating a new forecast that is an extension of a prior forecast.

        Args:
            statement_id: the ID of the statement to retrieve the most recent forecast for
            prediction_agent: the string for a prediction agent, if you want to further filter the most recent forecast

        Returns:
            int: the ID of the most recent forecast for the given statement
        """
        url = self.base_url + "get_most_recent_forecast/" + str(statement_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {}
        if prediction_agent is not None:
            data["prediction_agent"] = prediction_agent
        response = requests.post(url, headers=headers, json=data)
        if response.status_code == 200:
            return int(response.json()["forecast_id"])
        else:
            raise Exception(response.text)

    def get_forecast(self, forecast_id: int) -> dict:
        """
        Returns a specific forecast's details from the platform. This typically includes the forecast title, the value associated with the fill-in-the-blank component of a statement, and justificaiton for the forecast.

        Args:
            forecast_id: the ID of the forecast to retrieve

        Returns:
            dict: the forecast details in the form of a dictionary object
        """
        url = self.base_url + "get_forecast" + "/" + str(forecast_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(response.text)

    def create_forecast(
        self,
        statement_id: int,
        title: str,
        justification: str,
        value: float,
        prediction_agent: str,
        additional_data: dict = {},
        prior_forecast: int = None,
        is_human: bool = False,
    ) -> None:
        """
        Creates a forecast tied to a specific statement.

        Args:
            statement_id: the ID of the statement to tie the forecast to
            title: the title of the forecast
            justification: the justification for the forecast
            value: the value associated with the fill-in-the-blank component of the statement
            prediction_agent: the agent making the prediction
            additional_data: any additional data to include with the forecast. This is not used anywhere, but can be helpful in audting or researching forecast effectiveness
            prior_forecast: if this forecast is an extension of an earlier forecast, the ID of the prior forecast
            is_human: whether the prediction is human-generated

        Returns:
            dict: the newly created forecast from the platform
        """
        url = self.base_url + "create_forecast/" + str(statement_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "title": title,
            "justification": justification,
            "value": value,
            "prediction_agent": prediction_agent,
            "additional_data": additional_data,
            "is_human": is_human,
        }
        if prior_forecast is not None:
            data["prior_forecast"] = prior_forecast
        response = requests.post(url, headers=headers, data=json.dumps(data))
        if response.status_code == 201:
            return response.json()
        else:
            raise Exception(response.text)


class EmergingTrajectoriesClient(object):
    """
    This is the new and improved Emerging Trajectories client and should replace older clients. We keep the older ones for backwards compatability, for now.

    This client enables the same features as earlier clients, but also provides support for creating fact bases, automations, documents, and more.
    """

    # The base URL for the API, in case we need to change it or if someone wants to self-host anything.
    base_url = "https://emergingtrajectories.com/a/api/v0.2/"

    def __init__(self, api_key: str) -> None:
        """
        Args:
            api_key: the API key for the Emerging Trajectories platform.
        """
        self.api_key = api_key

    def create_factbase(self, title: str, description: str) -> str:
        """
        Create a new factbase on the Emerging Trajectories platform.

        Args:
            title: the title of the factbase
            description: a description of the factbase

        Returns:
            str: The short_code for the factbase.
        """
        url = self.base_url + "create_factbase"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "title": title,
            "description": description,
        }
        response = requests.post(url, headers=headers, data=json.dumps(data))
        if response.status_code == 200:
            r = response.json()
            if "short_code" in r:
                return r["short_code"]
        else:
            raise Exception(response.text)

    def create_document(
        self,
        factbase_shortcode: str,
        title: str = None,
        days_filter=99999,
        temperature=None,
        llm=None,
    ) -> int:
        """
        Create a new document on the Emerging Trajectories platform.

        Args:
            factbase_shortcode: the short code for the factbase to attach the document to
            title (optional): the title of the document
            days_filter (optional): limit facts to the recent number of days

        Returns:
            int: The ID of the document.
        """

        url = self.base_url + "api_doc_create"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "title": title,
            "factbase_shortcode": factbase_shortcode,
            "days_filter": days_filter,
        }

        if temperature is not None:
            data["temperature"] = temperature

        if llm is not None:
            data["llm"] = llm

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            if "doc_id" in r:
                return int(r["doc_id"])
        else:
            raise Exception(response.text)

        raise Exception("Failed to create document.")

    def get_document(self, doc_id: int) -> dict:
        """
        Get document data.

        Args:
            doc_id: the ID of the document

        Returns:
            int: The ID of the document.
        """

        url = self.base_url + "api_doc_get"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"doc_id": doc_id}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r
        else:
            raise Exception(response.text)

    def get_block(self, doc_id: int, block_named_id: str) -> str:
        """
        Get the content of a named block from a document.

        Args:
            doc_id: the ID of the document
            block_named_id: the named ID of the block to retrieve
        Returns:
            str: the content of the block
        """

        url = self.base_url + "api_block_get"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"doc_id": doc_id, "block_named_id": block_named_id}

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()["content"]
            return r
        else:
            raise Exception(response.text)

    def append_header_block(self, doc_id, text: str) -> bool:
        """
        Add a header block to a document.

        Args:
            doc_id: the ID of the document to append the header block to
            text: the text (header) to append
        Returns:
            bool: True if successful, False otherwise
        """

        url = self.base_url + "api_doc_append_header_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "text": text,
        }

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            return True

        return False

    def append_text_block(self, doc_id: int, text: str) -> bool:
        """
        Add a text block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            text: the text to append
        Returns:
            bool: True if successful, False otherwise
        """

        url = self.base_url + "api_doc_append_text_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "text": text,
        }

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            return True

        return False

    def append_askai_fts_block(
        self, doc_id: int, query: str, fts_term: str, named_id: str = None
    ) -> dict:
        """
        Add an AI (FTS) block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
            fts_term: the full text search (FTS) terms/query to use for the AI block
            named_id: the named ID of the AI block
        Returns:
            JSON dict with document information
        """
        url = self.base_url + "api_doc_append_askai_fts_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
            "arg": fts_term,
        }

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def append_risk_score_doc_block(
        self, doc_id: int, query: str, named_id: str = None
    ) -> dict:
        """
        Add a risk score block to a document. This risk score only evaluates based on what's in the document already; it does not query the fact base.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
            named_id: the named ID of the AI block
        Returns:
            JSON dict with document information
        """
        url = self.base_url + "api_doc_append_risk_score_doc_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
        }

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def append_askai_fts_cot_block(
        self, doc_id: int, query: str, fts_term: str, named_id: str = None
    ) -> dict:
        """
        Add an AI (FTS) block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
            fts_term: the full text search (FTS) terms/query to use for the AI block
            named_id: the named ID of the AI block
        Returns:
            JSON dict with document information
        """

        url = self.base_url + "api_doc_append_askai_fts_cot_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
            "arg": fts_term,
        }

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def append_askai_cot_block(
        self, doc_id: int, query: str, named_id: str = None
    ) -> dict:
        """
        Add an AI (COT) block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
            named_id: the named ID of the AI block
        Returns:
            JSON dict with document information
        """

        url = self.base_url + "api_doc_append_askai_cot_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
        }

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def append_askai_block(self, doc_id: int, query: str, named_id: str = None) -> dict:
        """
        Add an AI block to a document.

        Args:
            doc_id: the ID of the document to append the text block to
            query: the query to use for the AI block
        Returns:
            JSON dict with document information
        """

        url = self.base_url + "api_doc_append_askai_block"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "doc_id": doc_id,
            "query": query,
        }

        if named_id is not None:
            data["named_id"] = named_id

        response = requests.post(url, headers=headers, data=json.dumps(data))

        if response.status_code == 200:
            r = response.json()
            return r

        raise Exception(response.text)

    def create_automation_factbase(
        self, short_code, job_type, arg_string=None, args=None
    ):
        """
        Create an automation for a fact base.

        Args:
            short_code: the short code for the factbase
            job_type: the type of job to run
            arg_string: a string of arguments
            args: a dictionary of arguments if the job requires more than one

        Returns:
            dict: the automation object
        """

        url = self.base_url + "create_automation_for_factbase" + "/" + short_code
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "job_type": job_type,
        }
        if arg_string != None:
            data["arg_string"] = arg_string
        if "args" != None:
            data["args"] = args
        response = requests.post(url, headers=headers, data=json.dumps(data))

        j = response.json()
        return j["info"]

    def create_automation_document(self, doc_id, job_type, arg_string=None, args=None):
        """
        Create an automation for a document.

        Args:
            doc_id: the document ID
            job_type: the type of job to run
            arg_string: a string of arguments
            args: a dictionary of arguments if the job requires more than one

        Returns:
            dict: the automation object
        """

        url = self.base_url + "create_automation_for_documente" + "/" + str(doc_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {
            "job_type": job_type,
        }
        if arg_string != None:
            data["arg_string"] = arg_string
        if "args" != None:
            data["args"] = args
        response = requests.post(url, headers=headers, data=json.dumps(data))

        j = response.json()
        return j["info"]

    def get_factbase_automations(self, factbase_short_code):
        """
        Get an array of all automations that can be run on a factbase.

        Args:
            factbase_short_code: the short code for the factbase

        Returns:
            list: an array of automations that can be run on the factbase
        """
        url = self.base_url + f"get_automations/{factbase_short_code}"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        j = response.json()
        print(j)
        return j["automations"]

    def queue_automation(self, automation_id):
        """
        Queue an automation (i.e., request the Emerging Trajectories platform to run an automation).

        Args:
            automation_id: the ID of the automation to queue

        Returns:
            bool: True if successful, False otherwise
        """
        url = self.base_url + f"queue_automation/{automation_id}"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:
            return True
        return False

    def get_statement(self, statement_id: int) -> Statement:
        """
        Returns a given statement from the platform. Includes title, description, deadline, and fill-in-the-blank.

        Args:
            statement_id: the ID of the statement to retrieve

        Returns:
            Statement: the statement from the platform
        """
        url = self.base_url + "get_statement" + "/" + str(statement_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:
            r_obj = response.json()
            s = Statement(r_obj["title"], r_obj["fill_in_the_blank"])
            s.id = int(r_obj["id"])
            s.description = r_obj["description"]
            s.deadline = dateparser.parse(r_obj["deadline"])
            s.created_at = dateparser.parse(r_obj["created_at"])
            s.updated_at = dateparser.parse(r_obj["updated_at"])
            s.created_by = r_obj["created_by"]
            return s
        else:
            raise Exception(response.text)

    def get_forecast(self, forecast_id: int) -> Forecast:
        """
        Returns a given forecast from the platform.

        Args:
            forecast_id: the ID of the statement to retrieve

        Returns:
            Forecast: the forecast from the platform
        """
        url = self.base_url + "get_forecast" + "/" + str(forecast_id)
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(url, headers=headers)
        if response.status_code == 200:

            r_obj = response.json()

            f = Forecast(r_obj["title"], float(r_obj["value"]), r_obj["justification"])

            f.id = int(r_obj["forecast_id"])

            f.statement_id = int(r_obj["statement_id"])
            f.statement = self.get_statement(int(r_obj["statement_id"]))

            f.created_at = dateparser.parse(r_obj["created_at"])
            f.updated_at = dateparser.parse(r_obj["updated_at"])
            # f.created_by = r_obj["created_by"]
            f.prediction_agent = r_obj["prediction_agent"]

            f.additional_data = r_obj["additional_data"]

            if "prior_forecast" in r_obj:
                if r_obj["prior_forecast"] is not None:
                    f.prior_forecast = int(r_obj["prior_forecast"])
            f.is_human = bool(r_obj["is_human"])

            if "next_forecasts" in r_obj:
                if r_obj is not None:
                    f.next_forecasts = r_obj["next_forecasts"]

            return f
        else:
            raise Exception(response.text)

    def add_facts_to_factbase(
        self, fact_db_slug: str, url: str, facts: list[str]
    ) -> bool:
        """
        Adds a list of facts to a factbase on the Emerging Trajectories website.

        Args:
            fact_db_slug: the slug of the fact database to add the fact to.
            url: the URL of the fact.
            facts: the facts to add (a list of strings).

        Reutnr:
            bool: True if successful, False otherwise.
        """

        api_url = self.base_url + "add_facts/" + fact_db_slug
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        j = {
            "facts": facts,
            "url": url,
        }
        response = requests.post(api_url, headers=headers, json=j)

        if response.status_code == 200 or response.status_code == 201:
            return True
        print(response)
        return False

    def add_fact_to_factbase(self, fact_db_slug: str, url: str, fact: str) -> bool:
        """
        Adds a fact to a factbase on the Emerging Trajectories website.

        Args:
            fact_db_slug: the slug of the fact database to add the fact to.
            url: the URL of the fact.
            fact: the fact to add.

        Reutnr:
            bool: True if successful, False otherwise.
        """
        api_url = self.base_url + "add_fact/" + fact_db_slug
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        j = {
            "fact": fact,
            "url": url,
        }
        response = requests.post(api_url, headers=headers, json=j)

        if response.status_code == 200 or response.status_code == 201:
            return True
        print(response)
        return False

    def add_content_to_factbase(
        self, fact_db_slug: str, url: str, content: str, topic: str
    ) -> bool:
        """
        Sends content to the Emerging Trajectories website and extract facts from it.

        Args:
            fact_db_slug: the slug of the fact database to add the content to.
            url: the URL of the content. Note: we do not actually crawl this, we assume the content passed is the right conent.
            content: the content to extract facts from.
            topic: the topic of the content.

        Returns:
            bool: True if successful, False otherwise.
        """

        api_url = self.base_url + "add_content_to_factbase/" + fact_db_slug
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        j = {
            "content": content,
            "url": url,
            "topic": topic,
        }
        response = requests.post(api_url, headers=headers, json=j)

        if response.status_code == 200 or response.status_code == 201:
            return True
        print(response)
        return False



from .recursiveagent import ETClient
from .facts import FactBaseFileCache
from .utils import UtilityHelper
from . import Client, Statement, Forecast

from phasellm.llms import ChatBot, OpenAIGPTWrapper, ChatPrompt

from datetime import datetime

start_system_prompt = """Today's date is {the_date}. You are a researcher helping with economics and politics research. We will give you a few facts and we need you to fill in a blank to the best of your knowledge, based on all the information provided to you."""

start_user_prompt = """Here is the research:
---------------------
{content}
---------------------
{additional_facts}

Given the above, we need you to do your best to fill in the following blank...
{fill_in_the_blank}

PLEASE DO THE FOLLOWING:
- Provide any further justification ONLY BASED ON THE FACTS AND SOURCES PROVIDED ABOVE.
- Explain your forecast and how the facts, insights, etc. support it. Do not simply state a number.
- Do not provide a range; provide ONE number.
- End your forecast with the filled-in statement: {fill_in_the_blank_2}

We realize you are being asked to provide a speculative forecast. We are using this to better understand the world and finance, so please fill in the blank. We will not use this for any active decision-making, but more to learn about the capabilities of AI.
"""

extend_user_prompt = """Here is the research:
---------------------
{content}
---------------------
{additional_facts}

In addition to the new content above, we want to UPDATE the forecast from before. Here is the earlier forecast...
---------------------
FORECAST: {earlier_forecast_value}

JUSTIFICATION:
{earlier_forecast}
---------------------

Given the above, we need you to do your best to fill in the following blank...
{fill_in_the_blank}

PLEASE DO THE FOLLOWING:
- Provide any further justification ONLY BASED ON THE FACTS AND SOURCES PROVIDED ABOVE.
- Explain your forecast and how the facts, insights, etc. support it. Do not simply state a number.
- Do not provide a range; provide ONE number.
- End your forecast with the filled-in statement: {fill_in_the_blank_2}

We realize you are being asked to provide a speculative forecast. We are using this to better understand the world and finance, so please fill in the blank. We will not use this for any active decision-making, but more to learn about the capabilities of AI.

"""


class FactForecastingAgent(object):

    # TODO: document / clean up
    def __init__(
        self,
        client: ETClient,
        chatbot: ChatBot,
        factbase: FactBaseFileCache,
    ):

        self.client = client
        self.chatbot = chatbot
        self.factbase = factbase

    # TODO / NOTE: this allows us to continue chatting with the forecasting agent, since we can obtain the chatbot later. Given that some folks are interested in asking for clarifications, this could be an interesting opportunity.
    def setChatBot(self, chatbot):
        self.chatbot = chatbot

    # TODO: standardize -- camel case or snake case? Or something else?
    def getChatBot(self):
        return self.chatbot

    # TODO: we can do much better at disaggregating all these functions. Currently just want this to work.
    # TODO: Google query can be a list of queries, not just a single query.
    def create_forecast(
        self,
        statement: Statement,
        openai_api_key,
        et_api_key,
        google_api_key,
        google_search_id,
        google_search_query,
        facts=None,
        prediction_agent="Test Agent",
    ):

        fact_llm = OpenAIGPTWrapper(openai_api_key, "gpt-4-0125-preview")
        fact_chatbot = ChatBot(fact_llm)

        if isinstance(google_search_query, str):
            print("CALLING SINGLE QUERY...")
            content = self.factbase.summarize_new_info(
                statement,
                fact_chatbot,
                google_api_key,
                google_search_id,
                google_search_query,
            )
        elif isinstance(google_search_query, list):
            content = self.factbase.summarize_new_info_multiple_queries(
                statement,
                fact_chatbot,
                google_api_key,
                google_search_id,
                google_search_query,
            )
        else:
            raise ValueError(
                "google_search_query must be a string or a list of strings"
            )

        if content is None:
            print("No new content added to the forecast.")
            return None

        chatbot_messages = [
            {"role": "system", "content": start_system_prompt},
            {"role": "user", "content": start_user_prompt},
        ]

        chatbot = self.chatbot

        prompt_template = ChatPrompt(chatbot_messages)

        the_date = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")

        additional_facts = ""
        if facts is not None:
            additional_facts = "Some additional facts for consideration are below...\n"
            afctr = 1
            for f in facts:
                additional_facts += f"AF{afctr}: {f}\n"
                afctr += 1
            additional_facts += "---------------------\n\n"

        chatbot.messages = prompt_template.fill(
            statement_title=statement.title,
            statement_description=statement.description,
            statement_fill_in_the_blank=statement.fill_in_the_blank,
            fill_in_the_blank_2=statement.fill_in_the_blank,
            content=content,
            the_date=the_date,
            additional_facts=additional_facts,
        )

        assistant_analysis = chatbot.resend()

        print("\n\n\n")
        print(assistant_analysis)

        uh = UtilityHelper(openai_api_key)
        prediction = uh.extract_prediction(
            assistant_analysis, statement.fill_in_the_blank
        )

        client = Client(et_api_key)

        full_content = content + "\n\n-----------------\n\n" + assistant_analysis

        response = client.create_forecast(
            statement.id,
            "Prediction",
            full_content,
            prediction,
            prediction_agent,
            {
                "full_response_from_llm_before_source_cleanup": content,
                "full_response_from_llm": assistant_analysis,
                "extracted_value": prediction,
            },
        )

        return response

    def extend_forecast(
        self,
        forecast: Forecast,
        openai_api_key,
        et_api_key,
        google_api_key,
        google_search_id,
        google_search_query,
        facts=None,
        prediction_agent="Test Agent",
    ):

        fact_llm = OpenAIGPTWrapper(openai_api_key, "gpt-4-0125-preview")
        fact_chatbot = ChatBot(fact_llm)

        if isinstance(google_search_query, str):
            content = self.factbase.summarize_new_info(
                forecast.statement,
                fact_chatbot,
                google_api_key,
                google_search_id,
                google_search_query,
            )
        elif isinstance(google_search_query, list):
            content = self.factbase.summarize_new_info_multiple_queries(
                forecast.statement,
                fact_chatbot,
                google_api_key,
                google_search_id,
                google_search_query,
            )
        else:
            raise ValueError(
                "google_search_query must be a string or a list of strings"
            )

        if content is None:
            print("No new content added to the forecast.")
            return None

        chatbot_messages = [
            {"role": "system", "content": start_system_prompt},
            {"role": "user", "content": extend_user_prompt},
        ]

        chatbot = self.chatbot

        prompt_template = ChatPrompt(chatbot_messages)

        the_date = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")

        additional_facts = ""
        if facts is not None:
            additional_facts = "Some additional facts for consideration are below...\n"
            afctr = 1
            for f in facts:
                additional_facts += f"AF{afctr}: {f}\n"
                afctr += 1
            additional_facts += "---------------------\n\n"

        chatbot.messages = prompt_template.fill(
            statement_title=forecast.statement.title,
            statement_description=forecast.statement.description,
            statement_fill_in_the_blank=forecast.statement.fill_in_the_blank,
            fill_in_the_blank_2=forecast.statement.fill_in_the_blank,
            content=content,
            the_date=the_date,
            additional_facts=additional_facts,
            earlier_forecast_value=str(forecast.value),
            earlier_forecast=forecast.justification,
        )

        assistant_analysis = chatbot.resend()

        print("\n\n\n")
        print(assistant_analysis)

        uh = UtilityHelper(openai_api_key)
        prediction = uh.extract_prediction(
            assistant_analysis, forecast.statement.fill_in_the_blank
        )

        client = Client(et_api_key)

        full_content = content + "\n\n-----------------\n\n" + assistant_analysis

        response = client.create_forecast(
            forecast.statement.id,
            "Prediction",
            full_content,
            prediction,
            prediction_agent,
            {
                "full_response_from_llm_before_source_cleanup": content,
                "full_response_from_llm": assistant_analysis,
                "extracted_value": prediction,
            },
            forecast.id,
        )

        return response



from setuptools import setup, find_packages

from project_metadata import NAME, VERSION, AUTHOR, DESCRIPTION, LONG_DESCRIPTION

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author=AUTHOR,
    author_email="hello@phaseai.com",
    license="MIT",
    packages=find_packages(),
    install_requires=[
        "phasellm>=0.0.22",
        "Django>=5.0.0",
        "python-dotenv>=1.0.0",
        "dateparser>=1.2.0",
        "pytest-playwright",
        "beautifulsoup4",
        "chromadb",
        "feedparser",
        "pypdf",
        "faiss-cpu",
        "microsoft-bing-newssearch",
        "scrapingbee",
        "tiktoken",
    ],
    extras_require={
        "docs": [
            "furo",
            "sphinx>=7.1.2",
            "myst_parser>=2.0.0",
            "sphinx-autoapi>=2.1.1",
            "sphinx-autobuild>=2021.3.14",
        ]
    },
    python_requires=">=3.10.0",
    keywords="llm, nlp, ai, social, politics, economics",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
)



"""
Facts agent. Similar to knowledge agent but simply provides a list of facts and associated sources.

This abstracts away the fact generation from forecast creation, thus allowing us to test different prompting strategies and LLMs.
"""

import os
import json
import hashlib
import re

# Using JSONEncoder to be consistent with the Emerging Trajectories website and platform.
from django.core.serializers.json import DjangoJSONEncoder

from phasellm.llms import OpenAIGPTWrapper, ChatBot, ChatPrompt
from phasellm.agents import WebpageAgent, WebSearchAgent

from datetime import datetime

from . import Client
from .crawlers import crawlerPlaywright
from phasellm.llms import OpenAIGPTWrapper, ChatBot

# Number of search results to return from web searche (default value).
_DEFAULT_NUM_SEARCH_RESULTS = 10

facts_base_system_prompt = """You are a researcher tasked with helping forecast economic and social trends. The title of our research project is: {statement_title}.

The project description is as follows...
{statement_description}

We will provide you with content from reports and web pages that is meant to help with the above. We will ask you to review these documents, create a set of bullet points to inform your thinking. Rather than using bullet points, please list each as F1, F2, F3, etc... So that we can reference it.

The content we provided you contains source numbers in the format 'SOURCE: #'. When you extract facts, please include the citation in square brackets, with the #, like [#], but replace "#" with the actual Source # from the crawled content we are providing you.

For example, if you are referring to a fact that came under --- SOURCE: 3 ---, you would write something like: "Data is already trending to hotter temperatures [3]." Do not include the "#" in the brackets, just the number.

Thus, a bullet point would look like this:
F1: (information) [1]
F2: (information) [1]
F3: (information) [2]

... and so on, where F1, F2, F3, etc. are facts, and [1], [2] are the source documents you are extracting the facts from.
"""

facts_base_user_prompt = """Today's date is {the_date}. We will now provide you with all the content we've managed to collect. 

----------------------
{scraped_content}
----------------------

Please think step-by-step by (a) extracting critical bullet points from the above, and (b) share any insights you might have based on the facts.

The content we provided you contains source numbers in the format 'SOURCE: #'. When you extract facts, please include the citation in square brackets, with the #, like [#], but replace "#" with the actual Source # from the crawled content we are providing you.

For example, if you are referring to a fact that came under --- SOURCE: 3 ---, you would write something like: "Data is already trending to hotter temperatures [3]." Do not include the "#" in the brackets, just the actual number.

DO NOT PROVIDE A FORECAST, BUT SIMPLY STATE AND SHARE THE FACTS AND INSIGHTS YOU HAVE GATHERED.
"""


def uri_to_local(uri: str) -> str:
    """
    Convert a URI to a local file name. In this case, we typically will use an MD5 sum.

    Args:
        uri (str): The URI to convert.

    Returns:
        str: The MD5 sum of the URI.
    """
    uri_md5 = hashlib.md5(uri.encode("utf-8")).hexdigest()
    return uri_md5


# TODO Move to Utils.py, or elsewhere.
def clean_citations(assistant_analysis: str, ctr_to_source: dict) -> str:
    """
    The analysis currently contains numerical citations that are likely not in order, or in some cases are not used. We will update the cituations to follow the proper numerical order, and also include the URLs at the very end.

    Args:
        assistant_analysis: the analysis text from the assistant
        ctr_to_source: the mapping of citation number to source URL

    Returns:
        str: the cleaned analysis text, with citations following a proper numerical format and URIs at the end of the analysis
    """

    new_ctr_map = {}
    ctr = 1

    end_notes = "\n\n--- SOURCES ---\n\n"
    new_analysis = ""

    matches = re.finditer(r"\[\d+\]", assistant_analysis)

    last_index = 0
    for m in matches:

        m_start = m.start() + 1
        m_end = m.end() - 1

        old_ctr = int(m.group()[1:-1])
        uri = ctr_to_source[old_ctr]

        if old_ctr not in new_ctr_map:
            new_ctr_map[old_ctr] = ctr
            end_notes += f"{ctr}: {uri}\n"
            ctr += 1

        new_analysis += assistant_analysis[last_index:m_start] + str(
            new_ctr_map[old_ctr]
        )
        last_index = m_end

    if last_index != 0:
        new_analysis += assistant_analysis[last_index:] + end_notes

    else:
        new_analysis = assistant_analysis + end_notes + "No citations provided."

    return new_analysis


# TODO If this works, it should be an agent with setllm() supported, etc.
class FactBaseFileCache:

    def __init__(
        self, folder_path: str, cache_file: str = "cache.json", crawler=None
    ) -> None:
        """
        The KnowledgeBaseFileCache is a simple file-based cache for web content and local files. The cache stores the original HTML, PDF, or TXT content and tracks when (if ever) an agent actually accessed the content.

        Args:
            folder_path (str): The folder where the cache will be stored.
            cache_file (str, optional): The name of the cache file. Defaults to "cache.json".
        """
        self.root_path = folder_path
        self.root_parsed = os.path.join(folder_path, "parsed")
        self.root_original = os.path.join(folder_path, "original")
        self.cache_file = os.path.join(folder_path, cache_file)
        self.cache = self.load_cache()

        if crawler is None:
            self.crawler = crawlerPlaywright()
        else:
            self.crawler = crawler

    # TODO: this function is a new one compared to the KnowledgeBaseFileCache
    # TODO: refactor this + code where we run one query
    def summarize_new_info_multiple_queries(
        self,
        statement,
        chatbot,
        google_api_key,
        google_search_id,
        google_search_queries,
        fileout=None,
    ) -> str:

        self.google_api_key = google_api_key
        self.google_search_id = google_search_id
        self.google_search_queries = google_search_queries

        webagent = WebSearchAgent(api_key=self.google_api_key)

        scraped_content = ""
        added_new_content = False

        # We store the accessed resources and log access only when we successfully submit a forecast. If anything fails, we'll review those resources again during the next forecasting attempt.
        accessed_resources = []

        ctr = 0
        ctr_to_source = {}

        for google_search_query in self.google_search_queries:

            results = webagent.search_google(
                query=google_search_query,
                custom_search_engine_id=self.google_search_id,
                num=_DEFAULT_NUM_SEARCH_RESULTS,
            )

            added_new_content = False

            for result in results:
                if not self.in_cache(result.url):
                    ctr += 1
                    added_new_content = True

                    try:
                        page_content = self.get(result.url)
                        print(page_content)
                    except Exception as e:
                        print(f"Failed to get content from {result.url}\n{e}")
                        self.force_empty(result.url)
                        page_content = ""

                    accessed_resources.append(result.url)
                    # knowledge_base.log_access(result.url)

                    scraped_content += (
                        f"{page_content}\n\n--- SOURCE: {ctr}-------------------\n\n"
                    )
                    ctr_to_source[ctr] = result.url

        # We also check the knowledge base for content that was added manually.
        unaccessed_uris = self.get_unaccessed_content()
        for ua in unaccessed_uris:
            added_new_content = True
            ctr += 1
            page_content = self.get(ua)

            accessed_resources.append(ua)
            # knowledge_base.log_access(ua)

            scraped_content += (
                f"{page_content}\n\n--- SOURCE: {ctr}-------------------\n\n"
            )
            ctr_to_source[ctr] = ua

        if not added_new_content:
            print("No new content added to the forecast.")
            return None

        the_date = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")

        prompt_template = ChatPrompt(
            [
                {"role": "system", "content": facts_base_system_prompt},
                {"role": "user", "content": facts_base_user_prompt},
            ]
        )

        chatbot.messages = prompt_template.fill(
            statement_title=statement.title,
            statement_description=statement.description,
            statement_fill_in_the_blank=statement.fill_in_the_blank,
            scraped_content=scraped_content,
            the_date=the_date,
        )

        assistant_analysis = chatbot.resend()
        assistant_analysis_sourced = clean_citations(assistant_analysis, ctr_to_source)

        print("\n\n\n")
        print(assistant_analysis_sourced)

        if fileout is not None:
            with open(fileout, "w") as w:
                w.write(assistant_analysis_sourced)

        for ar in accessed_resources:
            self.log_access(ar)

        return assistant_analysis_sourced

    # TODO: this function is a new one compared to the KnowledgeBaseFileCache
    def summarize_new_info(
        self,
        statement,
        chatbot,
        google_api_key,
        google_search_id,
        google_search_query,
        fileout=None,
    ) -> str:

        self.google_api_key = google_api_key
        self.google_search_id = google_search_id
        self.google_search_query = google_search_query

        webagent = WebSearchAgent(api_key=self.google_api_key)
        results = webagent.search_google(
            query=self.google_search_query,
            custom_search_engine_id=self.google_search_id,
            num=_DEFAULT_NUM_SEARCH_RESULTS,
        )

        scraped_content = ""

        added_new_content = False

        # We store the accessed resources and log access only when we successfully submit a forecast. If anything fails, we'll review those resources again during the next forecasting attempt.
        accessed_resources = []

        ctr = 0
        ctr_to_source = {}

        for result in results:
            if not self.in_cache(result.url):
                ctr += 1
                added_new_content = True

                try:
                    page_content = self.get(result.url)
                    print(page_content)
                except Exception as e:
                    print(f"Failed to get content from {result.url}\n{e}")
                    self.force_empty(result.url)
                    page_content = ""

                accessed_resources.append(result.url)
                # knowledge_base.log_access(result.url)

                scraped_content += (
                    f"{page_content}\n\n--- SOURCE: {ctr}-------------------\n\n"
                )
                ctr_to_source[ctr] = result.url

        # We also check the knowledge base for content that was added manually.
        unaccessed_uris = self.get_unaccessed_content()
        for ua in unaccessed_uris:
            added_new_content = True
            ctr += 1
            page_content = self.get(ua)

            accessed_resources.append(ua)
            # knowledge_base.log_access(ua)

            scraped_content += (
                f"{page_content}\n\n--- SOURCE: {ctr}-------------------\n\n"
            )
            ctr_to_source[ctr] = ua

        if not added_new_content:
            print("No new content added to the forecast.")
            return None

        the_date = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")

        prompt_template = ChatPrompt(
            [
                {"role": "system", "content": facts_base_system_prompt},
                {"role": "user", "content": facts_base_user_prompt},
            ]
        )

        chatbot.messages = prompt_template.fill(
            statement_title=statement.title,
            statement_description=statement.description,
            statement_fill_in_the_blank=statement.fill_in_the_blank,
            scraped_content=scraped_content,
            the_date=the_date,
        )

        assistant_analysis = chatbot.resend()
        assistant_analysis_sourced = clean_citations(assistant_analysis, ctr_to_source)

        print("\n\n\n")
        print(assistant_analysis_sourced)

        if fileout is not None:
            with open(fileout, "w") as w:
                w.write(assistant_analysis_sourced)

        for ar in accessed_resources:
            self.log_access(ar)

        return assistant_analysis_sourced

    def save_state(self) -> None:
        """
        Saves the in-memory changes to the knowledge base to the JSON cache file.
        """
        with open(self.cache_file, "w") as f:
            json.dump(self.cache, f, cls=DjangoJSONEncoder)

    def load_cache(self) -> None:
        """
        Loads the cache from the cache file, or creates the relevant files and folders if one does not exist.
        """

        if not os.path.exists(self.root_path):
            os.makedirs(self.root_path)

        if not os.path.exists(self.root_parsed):
            os.makedirs(self.root_parsed)

        if not os.path.exists(self.root_original):
            os.makedirs(self.root_original)

        if not os.path.exists(self.cache_file):
            with open(self.cache_file, "w") as f:
                f.write("{}")

        with open(self.cache_file, "r") as f:
            return json.load(f)

    def in_cache(self, uri: str) -> bool:
        """
        Checks if a URI is in the cache already.

        Args:
            uri (str): The URI to check.

        Returns:
            bool: True if the URI is in the cache, False otherwise.
        """
        if uri in self.cache:
            return True
        return False

    def update_cache(
        self, uri: str, obtained_on: datetime, last_accessed: datetime
    ) -> None:
        """
        Updates the cache file for a given URI, specifically when it was obtained and last accessed.

        Args:
            uri (str): The URI to update.
            obtained_on (datetime): The date and time when the content was obtained.
            last_accessed (datetime): The date and time when the content was last accessed.
        """
        uri_md5 = uri_to_local(uri)
        self.cache[uri] = {
            "obtained_on": obtained_on,
            "last_accessed": last_accessed,
            "accessed": 0,
            "uri_md5": uri_md5,
        }
        self.save_state()

    def log_access(self, uri: str) -> None:
        """
        Saves the last accessed time and updates the accessed tracker for a given URI.

        Args:
            uri (str): The URI to update.
        """
        self.cache[uri]["last_accessed"] = datetime.now()
        self.cache[uri]["accessed"] = 1
        self.save_state()

    def get_unaccessed_content(self) -> list[str]:
        """
        Returns a list of URIs that have not been accessed by the agent.

        Returns:
            list[str]: A list of URIs that have not been accessed by the agent.
        """
        unaccessed = []
        for uri in self.cache:
            if self.cache[uri]["accessed"] == 0:
                unaccessed.append(uri)
        return unaccessed

    def force_empty(self, uri: str) -> None:
        """
        Saves an empty file for a given URI. Used when the page is erroring out.

        Args:
            uri (str): The URI to empty the cache for.
        """
        uri_md5 = uri_to_local(uri)

        with open(os.path.join(self.root_original, uri_md5), "w") as f:
            f.write("")
        with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
            f.write("")

        self.update_cache(uri, datetime.now(), datetime.now())

    def get(self, uri: str) -> str:
        """
        Returns the content for a given URI. If the content is not in the cache, it will be scraped and added to the cache.

        Args:
            uri (str): The URI to get the content for.

        Returns:
            str: The content for the given URI.
        """
        uri_md5 = uri_to_local(uri)
        if uri in self.cache:
            with open(os.path.join(self.root_parsed, uri_md5), "r") as f:
                return f.read()
        else:
            # scraper = WebpageAgent()

            # content_raw = scraper.scrape(uri, text_only=False, body_only=False)
            # with open(os.path.join(self.root_original, uri_md5), "w") as f:
            #    f.write(content_raw)

            # content_parsed = scraper.scrape(uri, text_only=True, body_only=True)
            # with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
            #    f.write(content_parsed)

            content, text = self.crawler.get_content(uri)
            with open(os.path.join(self.root_original, uri_md5), "w") as f:
                f.write(content)
            with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
                f.write(text)

            self.update_cache(uri, datetime.now(), datetime.now())

            return text

    def add_content(self, content: str, uri: str = None) -> None:
        """
        Adds content to cache.

        Args:
            content (str): The content to add to the cache.
            uri (str, optional): The URI to use for the content. Defaults to None, in which case an MD5 sum of the content will be used.
        """
        if uri is None:
            uri = hashlib.md5(content.encode("utf-8")).hexdigest()
        uri_md5 = uri_to_local(uri)
        with open(os.path.join(self.root_parsed, uri_md5), "w") as f:
            f.write(content)
        self.update_cache(uri, datetime.now(), datetime.now())

    def add_content_from_file(self, filepath: str, uri: str = None) -> None:
        """
        Adds content from a text file to the cache.

        Args:
            filepath (str): The path to the file to add to the cache.
            uri (str, optional): The URI to use for the content. Defaults to None, in which case an MD5 sum of the content will be used.
        """
        with open(filepath, "r") as f:
            content = f.read()
        self.add_content(content, uri)
